;module-definition file for Windows Zip DLL -- used by link.exe
LIBRARY ZIP32 ; Library module name
DESCRIPTION 'Windows Info-ZIP Zip Library 1.02 by Info-ZIP, Mike White 1997'

CODE  PRELOAD FIXED

DATA  PRELOAD MOVEABLE

EXPORTS
     ZpArchive
     ZpVersion
     ZpInit
     ZpSetOptions
     ZpGetOptions

