#include <windows.h>
#ifdef WIN32
#include <winver.h>
#endif
#include "zipver.h"

#ifndef WIN32
VS_VERSION_INFO VERSIONINFO
FILEVERSION 1,1,0,0
PRODUCTVERSION 3,0,0,0
FILEFLAGSMASK 0x3L
#ifndef MSC
FILEOS VOS__WINDOWS16
FILETYPE VFT_DLL
#else
FILEOS 0x1L //VOS__WINDOWS16
FILETYPE 0x2L //VFT_DLL
#endif
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904E4"
        BEGIN
            VALUE "CompanyName", COMPANY_NAME
            VALUE "FileDescription", "Info-ZIP's Zip dll\0"
            VALUE "FileVersion", ZIP_DLL_VERSION
            VALUE "InternalName", "Zip16\0"
            VALUE "LegalCopyright", "Info-ZIP 1997\0"
            VALUE "OriginalFilename", "ZIP16.DLL\0"
            VALUE "ProductName", "Info-ZIP's WiZ\0"
            VALUE "ProductVersion", ZIP_DLL_VERSION
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1252 
    END
END
#else
VS_VERSION_INFO VERSIONINFO
FILEVERSION 1,1,0,0
PRODUCTVERSION 3,0,0,0
FILEFLAGSMASK 0x3L
FILEOS VOS__WINDOWS32
FILETYPE VFT_DLL
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904E4"
        BEGIN
            VALUE "CompanyName", COMPANY_NAME
            VALUE "FileDescription", "Info-ZIP's Zip dll\0"
            VALUE "FileVersion", ZIP_DLL_VERSION
            VALUE "InternalName", "Zip32\0"
            VALUE "LegalCopyright", "Info-ZIP 1997\0"
            VALUE "OriginalFilename", "ZIP32.DLL\0"
            VALUE "ProductName", "Info-ZIP's WiZ\0"
            VALUE "ProductVersion", ZIP_DLL_VERSION
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1252 
    END
END
#endif

